VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsHID"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Option Explicit
'Project: HIDServer.vbp
'Version: 1.1
'Date: 3/8/00
'Copyright 1999 and 2000 by Jan Axelson (jan@lvr.com)
'
'Purpose:
'ActiveX Exe server that demonstrates background reading of reports from an HID-class
'USB device.

'Theory of operation:
'On request by an application, the server attempts to read Input reports from the
'specified device.
'When the server has read a report, it raises the ReportReady event in the
'requesting application.
'The application then calls the ReadReport routine in the server.
'The server returns the report data in the ReportData callback routine in the application.
'The server runs in its own thread, so it can wait for report data without
'blocking the application.
'On request, the server stops attempting to read from the device. This enables
'the server to close.

'Changes and updates:
'3/8/00
'To enable running under Windows 2000, the server now obtains a duplicate handle to the device
'instead of using the passed device pathname and CreateFile to open a handle.

Dim ApplicationHandle As Long
Dim DeviceHandle As Long
Dim DeviceHandleValid As Boolean
Dim DevicePathName As String
Dim HID As Long
Dim NumberOfBytesRead As Long
Dim ReportLength As Long
Dim ReadBuffer() As Byte
Dim ReadSuccess As Boolean

'The server raises the ReportReady event in the application
'when the server has read a new report from the device.
Event ReportReady()

Private Callbackobject As Object
'An instance of the form that holds the timer control.
Private frmServerTest As frmServer
'An instance of the timer on the form.
Private WithEvents tmrInitiateReadReport As Timer
Attribute tmrInitiateReadReport.VB_VarHelpID = -1

Private Function DoAReadFile() As Boolean
'Attempt to read a report.
Dim Result As Long

'******************************************************************************
'ReadFile
'Returns: the report in ReadBuffer.
'Requires: a device handle returned by CreateFile,
'the Input report length in bytes passed by the application (from HidP_GetCaps).
'******************************************************************************

'ReadFile is a blocking call. The thread will wait until the device
'sends the requested amount of data.

'Pass the address of the first byte of the read buffer.
Result = ReadFile _
    (HID, _
    ReadBuffer(0), _
    ReportLength, _
    NumberOfBytesRead, _
    0)

'Optional debugging information:
'Debug.Print "Readfile result= "; GetErrorString(Err.LastDllError)
'Debug.Print "handle= "; DeviceHandle
'Debug.Print "report length= "; ReportLength
'Debug.Print "bytes read= "; NumberOfBytesRead
'For Count = 0 To UBound(ReadBuffer)
'    Debug.Print ReadBuffer(Count)
'Next Count

If Result = 0 Then
    DoAReadFile = False
Else
    DoAReadFile = True
End If

End Function

Private Function GetErrorString _
    (ByVal LastError As Long) _
As String

'Returns the error message for the last error.
'Adapted from Dan Appleman's "Win32 API Puzzle Book"
Dim Bytes As Long
Dim ErrorString As String
ErrorString = String$(129, 0)
Bytes = FormatMessage _
    (FORMAT_MESSAGE_FROM_SYSTEM, _
    0&, _
    LastError, _
    0, _
    ErrorString$, _
    128, _
    0)
    
'Subtract two characters from the message to strip the CR and LF.
If Bytes > 2 Then
    GetErrorString = Left$(ErrorString, Bytes - 2)
End If

End Function

Private Sub Class_Initialize()
    Debug.Print "object created"
    DeviceHandleValid = False
    'Create an instance of the frmServer form and load it.
    Set frmServerTest = New frmServer
    Load frmServerTest
    'Create an instance of the form's timer.
    Set tmrInitiateReadReport = frmServerTest.tmrGenericTimer
End Sub

Private Sub Class_Terminate()
    Dim Result As Long
    Debug.Print "object terminated"
    'Close the handle to the device if necessary.
    'The pseudohandle to the server process doesn't have to be closed.
    Result = CloseHandle _
        (HID)
    Set tmrInitiateReadReport = Nothing
    Unload frmServerTest
    Set frmServerTest = Nothing
End Sub

Private Sub tmrInitiateReadReport_Timer()
'If the application has requested the server to read from the device, do so.
If DeviceHandleValid = True Then
    'Disable the timer to prevent new timer events while this routine is executing.
    tmrInitiateReadReport.Enabled = False
    'Attempt to read a report from the device.
    Call ReadReport(DevicePathName)
    '(The timer is re-enabled after the application has retrieved the report.)
End If
End Sub

Public Sub InitiateHIDReads _
    (ApplicationProcessID As Long, _
    ApplicationDeviceHID As Long, _
    ReadBufferPassed() As Byte, _
    Length As Long)
    
'The application calls this routine to tell the server to start attempting to
'read data from the device.

Dim PseudoHandle As Long
Dim Result As Long

'The passed parameters identify the device and describe the report.
'Save the parameters in module-level variables.
ReportLength = Length
ReadBuffer() = ReadBufferPassed()

'Get a handle to the application.
ApplicationHandle = OpenProcess _
    (PROCESS_VM_READ Or PROCESS_VM_WRITE, _
    False, _
    ApplicationProcessID)

'Get a handle to the device.
'Under Windows 2000, the server has to obtain a "duplicate handle" for the device.
'DuplicateHandle requires:
'Application Handle: the handle obtained from the passed Process ID.
'ApplicationDeviceHID: the device handle passed by the application.
'PseudoHandle: a handle for the current server process.
'DuplicateHandle returns:
'HID: a handle the server can use to access the device.

'Get a handle for this process:
PseudoHandle = GetCurrentProcess

'Get the duplicate handle for the device:
Result = DuplicateHandle _
    (ApplicationHandle, _
    ApplicationDeviceHID, _
    PseudoHandle, _
    HID, _
    0, _
    False, _
    DUPLICATE_SAME_ACCESS)

'Optional debugging information:
Debug.Print "Duplicate Handle = " & GetErrorString(Result)
Debug.Print "Application Handle = " & ApplicationHandle
Debug.Print "Application Device ID = " & ApplicationDeviceHID
Debug.Print "Server PseudoHandle = " & PseudoHandle
Debug.Print "HID: " & HID

'The application handle is no longer needed.
Result = CloseHandle _
        (ApplicationHandle)
        
'See if the device handle is valid.
'When DeviceHandleValid is true, the Timer-event routine attempts to read a report
'from the device, unless a previous read operation is in progress.
'Check for a valid device handle and exit if not present.
If HID = INVALID_HANDLE_VALUE Then
    'Handle is invalid, so don't attempt to read.
    DeviceHandleValid = False
Else
    'Handle is valid. OK to read from the device.
    DeviceHandleValid = True
End If

End Sub

Public Sub ReadReport(DevicePathName As String)
'If there is a valid device handle, attempt to read a report from the device.
Dim Count
Dim Handle As Long
Dim Result As Long
Dim UBoundReadBuffer As Integer

'The ReadBuffer array begins at 0, so subtract 1 from the number of bytes in the report.
If ReportLength > 0 Then
    ReDim ReadBuffer(ReportLength - 1)
End If

If DeviceHandleValid = True Then
    ReadSuccess = DoAReadFile
End If
  
'When the application wants to stop reading data, it sends an Output report to the device
'telling it to set byte 0 in its next Input report to FFh.
'Each time the server reads an Input report from the device, it checks byte 0.
'When the server receives a report with byte 0 = FFh, it closes the handle to the device
'and sets DeviceHandleValid false to stop attempting to read from the device in the timer event.
'Byte 0 in the report is byte 1 in the read buffer. (Byte 0 is the report index.)
If ReadBuffer(1) = &HFF Then
    'Close the handles to the device.
    Result = CloseHandle _
        (HID)
    DeviceHandleValid = False
End If

'Call the ReportReady routine in the application to let it know
'that the read operation is complete.
'Either a new report has been received or the read attempt failed.
RaiseEvent ReportReady
End Sub

Public Sub ReturnReportData _
    (callbackparam As Object, _
    ReadBufferparam() As Byte)

'The application calls this routine when it's been notified that a report is available.
'The server returns the last report read from the device
'by calling the ReportData routine in the application,
'passing it a byte array containing the report.

'Callbackparam is the object to call back (the application).
Set Callbackobject = callbackparam

'Return the report.
'ReportData is a routine in the application.
'Success indicates whether or not the read was successful.
Callbackobject.InputReportData _
    ReadSuccess, _
    ReadBuffer()
    
'Re-enable the timer so the server can read more reports.
tmrInitiateReadReport.Enabled = True
Set Callbackobject = Nothing
End Sub
